/*
 * dither.c:
 *
 * Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
 * All rights reserved.
 *
 */

static char rcsid[] =
  "$Id: dither.c,v 1.4 2001/05/25 01:24:37 root Exp $";

/*
 * $Log: dither.c,v $
 * Revision 1.4  2001/05/25 01:24:37  root
 * #
 *
 * Revision 1.3  2001/05/25 01:24:15  root
 * #
 *
 */


#include <stdio.h>
#include <unistd.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <getopt.h>
#include <string.h>
#include <strings.h>

extern void koch_dither (int w, int h, unsigned char *in, int xs, int ys,
                         unsigned char *out, int obpl);
extern void bfs_dither (int w, int h, unsigned char *in, int xs, int ys,
                        unsigned char *out, int obpl);
extern void thresh_dither (int w, int h, unsigned char *in, int xs, int ys,
                           unsigned char *out, int obpl);


static void
my_bcopy (unsigned char *in, unsigned char *out, int n)
{
  while (n--)
    *(out++) = *(in++);
}


static int
getfdline (int fd, char *buf, int buflen)
{
  char *ptr = buf;

  buflen--;

  while (buflen > 0)
    {

      read (fd, ptr, 1);
      *(ptr + 1) = 0;
      buflen--;

      if (*ptr == '\n')
        {
          if (*buf == '#')
            {
              ptr = buf;
            }
          else
            {
              return 1;
            }
        }
      else
        ptr++;
    }

  return 0;
}


int
main (int argc, char *argv[])
{
  char buf[1024];
  int infd, outfd;
  off_t inoff, outoff;
  size_t inlength, outlength;
  int format = -1;
  int w, h, m, obpl;
  int c;
  int koch = 0;
  int floyd = 0;
  int thresh = 0;

  int pagesize = getpagesize ();


  unsigned char *in, *inbase;
  unsigned char *out, *outbase;

  void (*dither) (int, int, unsigned char *, int, int, unsigned char *, int);

/* Two possible input types P5 or P6 */

  infd = 0;
  outfd = 1;

  while ((c = getopt (argc, argv, "i:o:kft")) != EOF)
    {
      switch (c)
        {
        case 'k':
          dither = koch_dither;
          koch++;
          break;
        case 'f':
          dither = bfs_dither;
          floyd++;
          break;
        case 't':
          dither = thresh_dither;
          thresh++;
          break;
        case 'i':
          infd = open (optarg, O_RDONLY);
          if (infd < 0)
            {
              perror ("couldn't open input file");
              exit (1);
            }
          break;
        case 'o':
          outfd = open (optarg, O_RDWR | O_CREAT | O_TRUNC, 0666);
          if (outfd < 0)
            {
              perror ("couldn't open output file");
              exit (1);
            }
          break;
        default:
          fprintf (stderr, "Usage:\n");
          fprintf (stderr, "%s {-f|-k|-t} [-i input] [-o output]\n", argv[0]);
          fprintf (stderr, "input and output default to stdin/stdout\n");
          exit (1);


        }

    }


  if ((koch + floyd + thresh) != 1)
    {
      fprintf (stderr, "Exactly one of -f,-k or -t  must be specified\n");
      exit (1);
    }


  getfdline (infd, buf, sizeof (buf));

  if (!strcmp (buf, "P5\n"))
    format = 5;
  if (!strcmp (buf, "P6\n"))
    format = 6;

  if (format < 0)
    {
      fprintf (stderr, "Input must be either PGM or PPM in raw format\n");
      exit (-1);
    }

  getfdline (infd, buf, sizeof (buf));
  m = w = h = 0;
  sscanf (buf, "%d %d", &w, &h);
  getfdline (infd, buf, sizeof (buf));
  sscanf (buf, "%d", &m);

  inoff = lseek (infd, 0L, SEEK_CUR);


  lseek (infd, 0L, SEEK_END);
  inlength = lseek (infd, 0L, SEEK_CUR);
  lseek (infd, 0L, SEEK_SET);

  fprintf (stderr,
           "Input: P%cM %d by %d pixel maxval %d, %d bytes long\n",
           (format == 5) ? 'G' : 'P', w, h, m, inlength);

  if (m != 255)
    {
      fprintf (stderr,
               "Code doesn't properly do maxval!=255 yet, expect trouble\n");
    }

  inlength += (pagesize - 1);
  inlength &= ~(pagesize - 1);

  inbase = mmap (NULL, inlength, PROT_READ, MAP_SHARED, infd, 0);

  if (inbase == MAP_FAILED)
    {
      fprintf (stderr, "Failed to mmap input file...\n");
      fprintf (stderr, "try using -i option\n");
      exit (1);
    }

  in = inbase + inoff;

  obpl = ((w + 7) >> 3);

  switch (format)
    {
    case 5:
      outoff = sprintf (buf, "P4\n%d %d\n", w, h);
      outlength = outoff + h * obpl;
      break;
    case 6:
      outoff = sprintf (buf, "P0\n%d %d\n", w, h);
      outlength = outoff + 3 * h * obpl;
      break;
    }

  lseek (outfd, outlength - 1, SEEK_SET);
  write (outfd, "", 1);
  lseek (outfd, 0L, SEEK_SET);

  fprintf (stderr, "Output: P%cM %d by %d pixel %d bytes long\n",
           (format == 5) ? 'B' : 'C', w, h, outlength);

  outlength += (pagesize - 1);
  outlength &= ~(pagesize - 1);

  outbase =
    mmap (NULL, outlength, PROT_READ | PROT_WRITE, MAP_SHARED, outfd, 0);

  if (outbase == MAP_FAILED)
    {
      fprintf (stderr, "Failed to mmap output file...\n");
      fprintf (stderr, "try using -o option\n");
      exit (1);
    }

  my_bcopy (buf, outbase, strlen (buf));
  out = outbase + outoff;



  switch (format)
    {
    case 5:

      dither (w, h, in, 1, w, out, obpl);
      break;
    case 6:
      dither (w, h, in, 3, 3 * w, out, obpl);
      out += h * obpl;
      in++;
      dither (w, h, in, 3, 3 * w, out, obpl);
      in++;
      out += h * obpl;
      dither (w, h, in, 3, 3 * w, out, obpl);
      break;
    }


  munmap (outbase, outlength);
  close (outfd);
  munmap (inbase, inlength);
  close (infd);

  return 0;
}
